from typing import Tuple

import qulacs_core

def create_observable_from_openfermion_file(file_path: str) -> qulacs_core.Observable: ...
def create_observable_from_openfermion_text(text: str) -> qulacs_core.Observable: ...
def create_split_observable(arg0: str) -> Tuple[qulacs_core.Observable,qulacs_core.Observable]: ...
